#include <sps2lib.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <sps2tags.h>

#include "PoolRoom.h"
#include "PlayingBall.h"
#include "GameRules.h"
#include "ModelController.h"
#include "GeomController.h"
#include <ode/ode.h>
#include "perftime.h"
#include "InputSystem.h"
#include "Camera.h"
#include "Debugger.h"
#include "Constants.h"
#include "CollisionManager.h"
#include "SoundSystem.h"
#include "RenderManager.h"
#include "HUD.h"
#include "TextureManager.h"
#include "GameManager.h"

#define MAX_CONTACTS 1
#define SCREEN_CLEAR_COLOUR 0x0, 0x0, 0x0 //R,G,B
#define DEFAULT_MICROPROGRAM "../render-0.5/lib/render_backface.bin"

int GAMESTATE;

GameRules        *rulesSet;
PoolRoom         *room;
ModelController  *modelController;
GeomController   *geomController;
InputSystem      *inputSystem;
Camera           *camera;
Debugger		 debug;
SoundSystem      *soundSystem;
RenderManager    *renderMgr;
HUD				 *hud;
TextureManager   *t;
GameManager      *gameManager;

dWorldID        world;
dSpaceID        space;
dJointGroupID   contactGroup;
dContact        contactArray[MAX_CONTACTS];
int             numContacts = 0;
float           dampener    = -0.25;
float           minPenDepth = 0.0005; //1/2 millimetre

//----------------------------------------------
// Function prototypes

void createODEWorld();
void drawTriangle();
void nearCallback(void *data, dGeomID o1, dGeomID o2);

//----------------------------------------------

void createODEWorld()
{
    world        = dWorldCreate();
    space        = dHashSpaceCreate(NULL);
    contactGroup = dJointGroupCreate(0);

    dWorldSetGravity(world, 0, -9.81, 0);
}

void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
	char  tmp[1024];
    float bouncyness = 0.0;

	// colliding two non-space geoms, so generate contact
	// points between o1 and o2
	// add these contact points to the simulation
	numContacts = dCollide (o1, o2, MAX_CONTACTS, &contactArray[0].geom, sizeof(dContact));

	if(numContacts > 0)
	{
		char* o1Type      = geomController->GetGeomType(o1);
		char* o2Type      = geomController->GetGeomType(o2);
	    bool  createJoint = true;
        float volume      = (100.0 * (contactArray[0].geom.depth - minPenDepth))/5 + 80;

		if (strcmp(o1Type, "Ball") == 0 && strcmp(o2Type, "Ball") == 0)
		{
			PlayingBall* bPtr1 = (PlayingBall*)geomController->GetGeomObject(o1);
			PlayingBall* bPtr2 = (PlayingBall*)geomController->GetGeomObject(o2);
            
            if(contactArray[0].geom.depth > minPenDepth)
                soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallBall), volume);

            bouncyness = 1.0;
            gameManager->ProcessCollision(bPtr1, bPtr2);
		}

		//Ball / playing surface collisions
		if (strcmp(o1Type, "Ball") == 0 && strcmp(o2Type, "PlayingSurface") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o1);
			const dReal *currentVel = dBodyGetLinearVel(bPtr->GetBodyID());
			dBodyAddForce(bPtr->GetBodyID(), dampener * currentVel[0], dampener * currentVel[1], dampener * currentVel[2]);
            bouncyness = 0.4;
        }

		if (strcmp(o1Type, "PlayingSurface") == 0 && strcmp(o2Type, "Ball") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o2);
			const dReal *currentVel = dBodyGetLinearVel(bPtr->GetBodyID());
			dBodyAddForce(bPtr->GetBodyID(), dampener * currentVel[0], dampener * currentVel[1], dampener * currentVel[2]);
            bouncyness = 0.4;
		}

		//Ball / room collisions. If you trigger these, you really suck.
		if (strcmp(o1Type, "Ball") == 0 && strcmp(o2Type, "Room") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o1);
			createJoint = false;
            
            soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallRoom));
			
            gameManager->ProcessCollision(bPtr, BallRoom);
		}

		if (strcmp(o1Type, "Room") == 0 && strcmp(o2Type, "Ball") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o2);
			createJoint = false;
            
            soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallRoom));
			
            gameManager->ProcessCollision(bPtr, BallRoom);
		}
		
		//Ball / pocket collisions
		if (strcmp(o1Type, "Ball") == 0 && strcmp(o2Type, "Pocket") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o1);
			createJoint = false;

            soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallPocket));
			
            gameManager->ProcessCollision(bPtr, BallPocket);
		}

		if (strcmp(o1Type, "Pocket") == 0 && strcmp(o2Type, "Ball") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o2);
			createJoint = false;
            
            soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallPocket), volume);

            gameManager->ProcessCollision(bPtr, BallPocket);
		}

		if (strcmp(o1Type, "Ball") == 0 && strcmp(o2Type, "Cushion") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o1);
            if(contactArray[0].geom.depth > minPenDepth)
                soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallCushion), volume);
            bouncyness = 0.8;
		}
		if (strcmp(o1Type, "Cushion") == 0 && strcmp(o2Type, "Ball") == 0)
		{
			PlayingBall* bPtr = (PlayingBall*)geomController->GetGeomObject(o2);
            if(contactArray[0].geom.depth > minPenDepth)
                soundSystem->PlaySound(rulesSet->GetSound(GameRules::BallCushion), volume);
            bouncyness = 0.8;
		}

        //If appropriate, do the collision joint stuff
		if(createJoint)
		{
            contactArray[0].surface.mode = dContactBounce;
			contactArray[0].surface.mu = 1;
			contactArray[0].surface.bounce = bouncyness;
			contactArray[0].surface.bounce_vel = 0.05;
			dJointID joint = dJointCreateContact(world, contactGroup, contactArray);
			dJointAttach(joint, dGeomGetBody(contactArray[0].geom.g1), dGeomGetBody(contactArray[0].geom.g2));
		}
	}
}

int main(int argc, char **argv)
{
	GAMESTATE   = GAMESTATE_INTRO;
    soundSystem = new SoundSystem();
	debug.Post("!STARTED");
   
	int sps2descriptor = sps2Init();    // initialise sps2

	if (sps2descriptor < 0) {
        printf("Unable to initialise SPS2.\n");
        exit(-1);
    }

	debug.Post("SPS2 Initialised.");

    sps2UScreenInit(0);
    Render_AllocateBuffers(sps2descriptor);
    void *microcodeaddr = Render_LoadMicrocode(DEFAULT_MICROPROGRAM);	// load microcode into data buffer
    Render_AddCallTag(microcodeaddr);
    void *buffersetupaddr = Render_SetupVUBuffers(0, 512);        		// set up VIF for double buffering
    Render_AddCallTag(buffersetupaddr);

	debug.Post("Calculating Lights...");

    lightinfo lights;                               // set parallel lights
    lights.direction[0] = Vec3(1.0f, 1.0f, 0.0f);
    lights.direction[1] = Vec3(1.0f, 1.0f, 0.0f);
    lights.direction[2] = Vec3(1.0f, 0.0f, 1.0f);
    lights.intensity[0] = 80;
    lights.intensity[1] = 50;
    lights.intensity[2] = 50;
    lights.ambient      = 16;
	
	debug.Post("Starting the Physics Engine");

    createODEWorld();

    rulesSet        = new GameRules();
    renderMgr       = new RenderManager(sps2descriptor);
    modelController = new ModelController(renderMgr);
	t               = new TextureManager(sps2descriptor);
 	geomController  = new GeomController();
    room            = new PoolRoom(rulesSet, world, space, modelController, geomController);
    camera          = new Camera();
    gameManager     = new GameManager(rulesSet, room, camera);
	hud				= new HUD(t, gameManager, sps2descriptor);
    inputSystem     = new InputSystem(room, camera, soundSystem, rulesSet, hud, gameManager);

    debug.SetGameRules(rulesSet);
    debug.SetPoolRoom(room);
	debug.SetCamera(camera);

	rulesSet->LoadSounds(soundSystem);
    room->GetUtensil()->SetAnimationLength(60);
    room->GetUtensil()->SetStrikeLength(10);

    init_perftime();
    
    __u64 oldtimer;
    __u64 timercount = 0;
    float cspeed     = (float)get_clock() * 1000000;
    float deltaTime, fps;
    Mat44 worldToScreen;

    float camerax = 0.0, cameray = 0.0, cameraz = 0.0;

	debug.Post("Starting Main Game Loop...");
    while(GAMESTATE != GAMESTATE_EXITING) {				// render loop
        soundSystem->Update();
        poll_counters();
        oldtimer      = timercount;
        timercount    = read_counter0();
        deltaTime     = (float)(timercount - oldtimer) / cspeed;

        inputSystem->CheckInput(GAMESTATE);          //Handle player input
		if(debug.OhPlease() && GAMESTATE == GAMESTATE_PLAYING)
        {
            // collide all objects together
            dSpaceCollide(space, 0, &nearCallback);

            //Do the physics
            dWorldStep(world, 0.01);//deltaTime / 2.0); //THIS FUNCTION IS THE CAUSE OF THE SLOWDOWN

			//Clear the collision joints
			dJointGroupEmpty(contactGroup);

            room->UpdateBallVelocities();
        }

		switch(GAMESTATE)
        {
            case GAMESTATE_EXITING:
                break;

            case GAMESTATE_MENU:
                break;

            case GAMESTATE_SHOWRESULTS:
                break;

            case GAMESTATE_PLAYING:
                if(room->AllBallsStopped())
                {
			        GAMESTATE = GAMESTATE_WAITING;
                    gameManager->EndTurn();
                    if(gameManager->IsGameOver())
                        GAMESTATE = GAMESTATE_SHOWRESULTS;
                }
                break;

            case GAMESTATE_WAITING:
                camerax = room->GetWhiteBall()->GetXPos();
                cameray = room->GetWhiteBall()->GetYPos();
                cameraz = room->GetWhiteBall()->GetZPos();
                break;

            case GAMESTATE_ANIMATECUE:
                room->GetUtensil()->DoAnimationStep();
                if(room->GetUtensil()->AnimationIsComplete())
				{
					soundSystem->PlaySound(rulesSet->GetSound(GameRules::CueBall), 255);
                    GAMESTATE = GAMESTATE_PLAYING;
				}
                break;
        }
        
		if((GAMESTATE != GAMESTATE_INTRO) && (GAMESTATE != GAMESTATE_OUTRO) && (GAMESTATE != GAMESTATE_EXITING))
		{
			worldToScreen = camera->MakeWorldToScreenMatrix(camerax, cameray, cameraz);
			room->Draw(worldToScreen, lights);
                        room->target->Draw(worldToScreen, lights);
                        room->targetPocket->Draw(worldToScreen, lights);
                        room->targetBall->Draw(worldToScreen, lights);
                        room->startingPoint->Draw(worldToScreen, lights);
			if((GAMESTATE == GAMESTATE_WAITING) || (GAMESTATE == GAMESTATE_ANIMATECUE))
				room->GetUtensil()->DrawUtensil(worldToScreen, lights, camera->GetYRotation(), camerax, cameray, cameraz);
		}

        soundSystem->Update();
        renderMgr->FinishedFrame();

		if(GAMESTATE != GAMESTATE_EXITING)
            hud->drawHUD(GAMESTATE, debug.SockNo);
        
        sps2UScreenSwap();	               				// swap buffers and clear new draw buffer
        sps2UScreenClear(SCREEN_CLEAR_COLOUR);
    }
	debug.Post("Closing down the game.");
    close_perftime();
    delete gameManager;
    delete soundSystem;
    Render_FreeBuffers();
    delete inputSystem;
    delete t;
	delete geomController;
    delete rulesSet;
    delete renderMgr;
    delete camera;
    delete hud;
	debug.Post("All Classes destroyed.");
    dWorldDestroy (world);
	dCloseODE();
    sps2UScreenShutdown();
    sps2Release(sps2descriptor);        					// shut down sps2
	debug.Post("SPS2 has been shutdown.");
    delete room;
    delete modelController;
}
